@@ -304,7 +304,7 @@ static LXSession *lxsession_add(void)
 
 static LXSession *lxsession_greeter(void)
 {
-	char temp[16];
+	char temp[128];
 	LXSession *s;
 	s=lxsession_find_greeter();
 	if(s)
@@ -327,10 +327,17 @@ static LXSession *lxsession_greeter(void)
 	s->idle=FALSE;
 	sprintf(temp,":%d",s->display);
 	setenv("DISPLAY",temp,1);
+	#ifndef DISABLE_XAUTH
+	sprintf(temp,"/var/run/lxdm/lxdm-:%d.auth",s->display);
+	setenv("XAUTHORITY",temp,1);
+	#endif
 	g_message("prepare greeter on %s\n",temp);
 	ui_prepare();
 	lxsession_set_active(s);
 	g_message("start greeter on %s\n",temp);
+	#ifndef DISABLE_XAUTH
+	unsetenv("XAUTHORITY");
+	#endif
 	return s;
 }
 
@@ -468,7 +475,7 @@ static char *lxsession_xserver_command(LXSession *s)
 		}
 	}
 
-	arg = g_renew(char *, arg, arc + 10);
+	arg = g_renew(char *, arg, arc + 15);
 	if(nr_tty)
 	{
 		arg[arc++] = g_strdup("-background");
@@ -486,6 +493,10 @@ static char *lxsession_xserver_command(LXSession *s)
 	{
 		arg[arc++] = g_strdup("-novtswitch");
 	}
+#ifndef DISABLE_XAUTH
+	arg[arc++] = g_strdup("-auth");
+	arg[arc++] = g_strdup_printf("/var/run/lxdm/lxdm-:%d.auth",s->display);
+#endif
 	arg[arc] = NULL;
 	p=g_strjoinv(" ", arg);
 	g_strfreev(arg);
@@ -694,7 +705,7 @@ static void create_server_auth(LXSession *s)
 
 	authfile = g_strdup_printf("/var/run/lxdm/lxdm-:%d.auth",s->display);
 
-	//setenv("XAUTHORITY",authfile,1);
+	setenv("XAUTHORITY",authfile,1);
 	remove(authfile);
 	xauth_write_file(authfile,s->display,s->mcookie);
 	g_free(authfile);
@@ -886,26 +897,57 @@ static void put_lock(void)
     g_free(lockfile);
 }
 
+static int get_run_level(void)
+{
+#if defined(HAVE_UTMPX_H) && defined(RUN_LVL)
+	int res;
+	struct utmpx *ut,tmp;
+
+	setutxent();
+	tmp.ut_type=RUN_LVL;
+	ut=getutxid(&tmp);
+	if(!ut)
+	{
+		endutxent();
+		return '5';
+	}
+	res=ut->ut_pid & 0xff;
+	endutxent();
+	//g_message("runlevel %c\n",res);
+	return res;
+#else
+	return '5';
+#endif
+}
+
 static void on_xserver_stop(void *data,int pid, int status)
 {
 	LXSession *s=data;
 	LXSession *greeter;
-
-	g_message("xserver stop, restart. return status %x\n",status);
+	int level;
 
 	stop_pid(pid);
 	s->server = -1;
 	lxsession_stop(s);
+	
+	level=get_run_level();
+	if(level=='6' || level=='0')
+	{
+		return;
+	}
+	
+	g_message("xserver stop, restart. return status %x\n",status);
+
 	greeter=lxsession_find_greeter();
 	if(s->greeter || !greeter)
 	{
 		s->greeter=TRUE;
 		xconn_close(s->dpy);
 		s->dpy=NULL;
-		lxdm_startx(s);
 		ui_drop();
+		lxdm_startx(s);
 		ui_prepare();
-		lxsession_set_active(greeter);
+		lxsession_set_active(s);
 	}
 	else
 	{
@@ -914,7 +956,7 @@ static void on_xserver_stop(void *data,int pid, int status)
 	}
 }
 
-void lxdm_startx(LXSession *s)
+static void lxdm_startx(LXSession *s)
 {
 	char *arg;
 	char **args;
@@ -967,6 +1009,9 @@ void lxdm_startx(LXSession *s)
 	g_message("%ld: start xserver in %d retry",time(NULL),i);
 	if(s->dpy==NULL)
 		exit(EXIT_FAILURE);
+	#ifndef DISABLE_XAUTH
+	unsetenv("XAUTHORITY");
+	#endif
 	
 	if(s->option && g_key_file_has_key(config,s->option,"numlock",NULL))
 	{
@@ -997,27 +1042,32 @@ static void exit_cb(void)
 	g_key_file_free(config);
 }
 
-static int get_run_level(void)
+static gboolean delayed_restart_greeter(LXSession *s)
 {
-#if defined(HAVE_UTMPX_H) && defined(RUN_LVL)
-	int res=0;
-	struct utmpx *ut,tmp;
-
-	setutxent();
-	tmp.ut_type=RUN_LVL;
-	ut=getutxid(&tmp);
-	if(!ut)
+	int level;
+	
+	level=get_run_level();
+	if(level=='0' || level=='6')
 	{
-		endutxent();
-		return 5;
+		if(level=='0')
+			g_spawn_command_line_sync("/etc/lxdm/PreShutdown",0,0,0,0);
+		else
+			g_spawn_command_line_sync("/etc/lxdm/PreReboot",0,0,0,0);
+		g_message("run level %c\n",level);
+		lxdm_quit_self(0);
+		return FALSE;
 	}
-	res=ut->ut_pid & 0xff;
-	endutxent();
-	//g_message("runlevel %c\n",res);
-	return res;
-#else
-	return 5;
-#endif
+	
+	if(s && s!=lxsession_greeter())
+	{
+		lxsession_free(s);
+	}
+	else if(!s)
+	{
+		lxsession_greeter();
+	}
+	
+	return FALSE;
 }
 
 static void on_session_stop(void *data,int pid, int status)
@@ -1047,15 +1097,9 @@ static void on_session_stop(void *data,int pid, int status)
 			g_spawn_command_line_sync("/etc/lxdm/PreReboot",0,0,0,0);
 		g_message("run level %c\n",level);
 		lxdm_quit_self(0);
+		return;
 	}
-	if(s && s!=lxsession_greeter())
-	{
-		lxsession_free(s);
-	}
-	else if(!s)
-	{
-		lxsession_greeter();
-	}
+	g_timeout_add(300,(GSourceFunc)delayed_restart_greeter,s);
 }
 
 gboolean lxdm_get_session_info(const char *session,char **pname,char **pexec,char **pdesktop_names)
@@ -1269,6 +1313,9 @@ void lxdm_do_login(struct passwd *pw, char *session, char *lang, char *option)
 	{
 		if(s) lxsession_free(s);
 		lxsession_set_active(prev);
+		g_free(session_name);
+		g_free(session_exec);
+		g_free(session_desktop_names);
 		return;
 	}
 	if(!s) s=lxsession_find_idle();
@@ -1514,7 +1561,7 @@ static void lxdm_signal_handler(void *data,int sig)
 	switch(sig){
 	case SIGTERM:
 	case SIGINT:
-		g_critical("QUIT BY SIGNAL\n");
+		g_critical("QUIT BY SIGNAL %d\n",sig);
 		lxdm_quit_self(0);
 		break;
 	default:
@@ -1650,6 +1697,7 @@ static GString *lxdm_user_cmd(void *data,int user,int arc,char **arg)
 		if(p)
 		{
 			res=g_string_new_len(p,len);
+			g_free(p);
 		}
 		g_key_file_free(kf);
 	}
